"""
drugstone.scripts.request_task_info

This module implements the request_task_info function.

:copyright: 2022 Institute for Computational Systems Biology by Prof. Dr. Jan Baumbach
:author: Ugur Turhan
"""

import requests
from .constants.url import Url


def request_task_info(token: str) -> dict:
    return requests.get(
        Url.TASK + "?token=" + token,
        verify=False).json()["info"]



"""
drugstone.task.task_result

This module implements the class :class:`TaskResult` for the drugstone API.

:copyright: 2022 Institute for Computational Systems Biology by Prof. Dr. Jan Baumbach
:author: Ugur Turhan
"""

import json
from pandas.core.frame import DataFrame
from .scripts.download_network_graph import download_network_graph
from .scripts.create_path import create_path


class TaskResult:
    """Represents the results of a task.

    get_genes() -> dict:
        Returns a dict with the genes.

    get_drugs() -> dict:
        Returns a dict with the drugs.

    to_dict() -> dict:
        Returns a dict with the result.

    to_pandas_dataframe() -> DataFrame:
        Returns a pandas :class:`DataFrame` of the result.

    download_json(path: str, name: str) -> None:
        Downloads a json file with the result.

    download_genes_csv(path: str, name: str) -> None:
        Downloads a csv file with the genes of the result.

    download_drugs_csv(path: str, name: str) -> None:
        Downloads a csv file with the drugs of the result.

    download_edges_csv(path: str, name: str) -> None:
        Downloads a csv file with the edges of the result.

    download_graph(path: str, name: str) -> None:
        Downloads a html file with a graph of the nodes.
    """

    def __init__(
            self,
            drugs: dict = dict({}),
            genes: dict = dict({}),
            raw_data: dict = dict({})) -> None:
        self.__drugs = drugs
        self.__genes = genes
        self.__raw_data = raw_data

    def get_genes(self) -> dict:
        """Returns a dict with the genes."""

        return self.__genes

    def get_drugs(self) -> dict:
        """Returns a dict with the drugs."""

        return self.__drugs

    def get_raw_result(self) -> dict:
        """Returns the raw data of the task."""
        return self.__raw_data

    def to_dict(self) -> dict:
        """Returns a dict with the result."""

        return {"drugs": self.__drugs, "genes": self.__genes}

    def to_pandas_dataframe(self) -> DataFrame:
        """Returns a pandas :class:`DataFrame` of the result."""

        return DataFrame(self.to_dict()).T

    def download_json(self, path: str = "", name: str = "result") -> None:
        """Downloads a json file with the result.

        :param str path: (optional) Path, where to download the file. Defaults to the current path.
        :param str name: (optional) Name for the file. Defaults to 'result'.
        """

        path = create_path(path, name, "json")
        with open(path, "x") as f:
            json.dump(self.to_dict(), f, indent=4)

    def download_raw_json(self, path: str = "", name: str = "raw_data") -> None:
        """Downloads a json file with the unprocessed results,
        as they are received from the backend.

        :param str path: (optional) Path, where to download the file. Defaults to the current path.
        :param str name: (optional) Name for the file. Defaults to 'raw_data'.
        """

        path = create_path(path, name, "json")
        with open(path, "x") as f:
            json.dump(self.get_raw_result(), f, indent=4)

    def download_genes_csv(self, path: str = "", name: str = "genes") -> None:
        """Downloads a csv file with the genes of the result.

        :param str path: (optional) Path, where to download the file. Defaults to the current path.
        :param str name: (optional) Name for the file. Defaults to 'genes'.
        """

        downloads_path = create_path(path, name, "csv")
        df = DataFrame(self.get_genes()).T
        df.to_csv(downloads_path)

    def download_drugs_csv(self, path: str = "", name: str = "drugs") -> None:
        """Downloads a csv file with the drugs of the result.

        :param str path: (optional) Path, where to download the file. Defaults to the current path.
        :param str name: (optional) Name for the file. Defaults to 'drugs'.
        """

        downloads_path = create_path(path, name, "csv")
        df = DataFrame(self.get_drugs()).T
        df.to_csv(downloads_path)

    def download_edges_csv(self, path: str = "", name: str = "edges") -> None:
        """Downloads a csv file with the edges of the result.

        :param str path: (optional) Path, where to download the file. Defaults to the current path.
        :param str name: (optional) Name for the file. Defaults to 'edges'.
        """

        edges = []
        for g in self.get_genes():
            for e in self.get_genes()[g]["has_edges_to"]:
                edges.append([g, e])
        data = {}
        for g in self.get_genes():
            data[g] = {}
            for e in self.get_genes():
                data[g][e] = 1 if [g, e] in edges or [e, g] in edges else 0
        df = DataFrame(data)
        downloads_path = create_path(path, name, "csv")
        df.to_csv(downloads_path)

    def download_graph(self, path: str = "", name: str = "graph") -> None:
        """Downloads a html file with a graph of the nodes.

        :param str path: (optional) Path, where to download the file. Defaults to the current path.
        :param str name: (optional) Name for the file. Defaults to 'graph'.
        """

        download_network_graph(self.to_dict(), path=path, name=name)



"""
drugstone.scripts.start_task

This module implements the start_task function.

:copyright: 2022 Institute for Computational Systems Biology by Prof. Dr. Jan Baumbach
:author: Ugur Turhan
"""

import requests
from .constants.url import Url


def start_task(params: dict) -> str:
    """Starts a task and returns the token for it.

    If there is no token for any reason, an empty str gets returned.
    """

    start_task_response = requests.post(
        Url.TASK,
        verify=False,
        json=params
    )

    token_obj = start_task_response.json()

    return token_obj.get("token", "")



"""
drugstone.task.scripts.make_upsetplot

This module implements the make_upset_plot function.

:copyright: 2022 Institute for Computational Systems Biology by Prof. Dr. Jan Baumbach
:author: Ugur Turhan
"""

from upsetplot import from_contents, plot
from matplotlib import pyplot


def make_upset_plot(tasks: dict):
    """Opens a new window with an upset-plot of the results."""

    data = {}
    for t in tasks:
        data[t] = list(tasks[t]["results"]["drugs"].keys())

    d = from_contents(data)
    plot(d)
    pyplot.show()



"""
drugstone.scripts.normalize_nodes

This module implements the normalize_nodes function.

:copyright: 2022 Institute for Computational Systems Biology by Prof. Dr. Jan Baumbach
:author: Ugur Turhan
"""


def normalize_nodes(results: dict) -> dict:
    """Returns a normalized dict of the drugs and genes."""

    drugs = {}
    genes = {}

    # For all the nodes in the results,
    # differentiates if it's a drug or gene
    # and puts it in the according dict.
    nodes = results["nodeAttributes"]["details"]
    node_types = results["nodeAttributes"]["nodeTypes"]
    for n_id, n_type in node_types.items():
        nodes[n_id]["node_type"] = n_type
    for _, node in nodes.items():
        if node["node_type"] == "drug":
            n_name = node["label"]
            drugs[n_name] = node
        elif node["node_type"] == "protein":
            n_name = node["symbol"]
            genes[n_name] = node

    # Adds to the genes if it's a seed or not.
    is_seed = results["nodeAttributes"]["isSeed"]
    for _, g_details in genes.items():
        g_details["is_seed"] = is_seed[g_details["netexId"]]

    # Normalizes the scores for the drugs.
    drug_scores = []
    for _, drug in drugs.items():
        if "score" in drug:
            drug_scores.append(drug["score"])
    if drug_scores:
        max_drug_score = max([x for x in drug_scores if x is not None])
    for _, drug in drugs.items():
        if "score" in drug and (type(drug["score"]) == int
                                or type(drug["score"]) == float):
            old_score = drug["score"]
            new_score = round(old_score / max_drug_score, 4)
            drug["score"] = new_score
        else:
            drug["score"] = None

    # sorts drugs without a score out
    none_score_drugs = []
    for drug, detail in drugs.items():
        if detail["score"] is None:
            none_score_drugs.append(drug)
    for x in none_score_drugs:
        drugs.pop(x)

    # Normalizes the scores for the genes.
    gene_scores = [1]
    for _, gene in genes.items():
        if "score" in gene:
            gene_scores.append(gene["score"])
    max_gene_score = max([x for x in gene_scores if x is not None])
    for _, gene in genes.items():
        if "score" in gene and (type(gene["score"]) == int
                                or type(gene["score"]) == float):
            old_score = gene["score"]
            new_score = round(old_score / max_gene_score, 4)
            gene["score"] = new_score
        else:
            gene["score"] = None

    # Adds the edges to the genes.
    edges = results["network"]["edges"]
    for _, gene in genes.items():
        edges_dict = [x for x in edges if x["from"] == gene["netexId"]]
        edges_netex_id = []
        edges_normalized = []
        for e in edges_dict:
            edges_netex_id.append(e["to"])
        for e in edges_netex_id:
            if str(e).startswith("p"):
                for _, g in genes.items():
                    if e == g["netexId"]:
                        edges_normalized.append(g["symbol"])
            elif str(e).startswith("d"):
                for _, d in drugs.items():
                    if e == d["netexId"]:
                        edges_normalized.append(d["label"])
            else:
                edges_normalized.append(e)
        gene["has_edges_to"] = edges_normalized

    # Removes unnecessary properties from drugs.
    for _, drug in drugs.items():
        drug.pop("netexId")
        drug.pop("trialLinks")
        drug.pop("node_type")

    # Removes unnecessary properties from genes.
    for _, gene in genes.items():
        gene.pop("netexId")
        gene.pop("node_type")

    return {"drugs": drugs, "genes": genes}



"""
drugstone.scripts.check_result_size

This module implements the check_result_size function.

:copyright: 2022 Institute for Computational Systems Biology by Prof. Dr. Jan Baumbach
:author: Ugur Turhan
"""


def check_result_size(result: dict, parameters: dict) -> dict:
    """Returns the results according to the result_size.

    Only edits the targeted results,
    e.g. for a drug-search, only the drugs will be edited.

    :param dict result: Dictionary of the result.
    :param dict parameters: Dictionary of the task parameters.
    :return: Dictionary of the results, according to the result_size.
    """

    target = parameters["target"]
    result_size = parameters["resultSize"]
    algorithm = ["algorithm"]

    drugs = result["drugs"]
    genes = result["genes"]

    if target == "drug":
        if len(drugs) <= result_size:
            return {"drugs": drugs, "genes": genes}
        if len(drugs) > result_size:
            sorted_drugs_list = sorted(drugs.values(), key=lambda item: item['score'])
            if algorithm != "proximity":
                # drugs are sorted from low to high score
                # usually the highest scored drugs are wanted
                # so sorted_drugs is reversed
                # only with the proximity algorithm
                # the lowest scored drugs are wanted
                sorted_drugs_list.reverse()
            resized_drugs = {}
            for drug in sorted_drugs_list[0:result_size]:
                label = drug["label"]
                resized_drugs[label] = drug
            drug_filtered_genes = {}
            for gene, detail in genes.items():
                old_edges = detail["has_edges_to"]
                new_edges = []
                for edge in old_edges:
                    if edge in resized_drugs.keys():
                        new_edges.append(edge)
                drug_filtered_genes[gene] = {
                    **detail,
                    "has_edges_to": new_edges
                }
            all_edges = []
            for _, gene in drug_filtered_genes.items():
                for e in gene["has_edges_to"]:
                    all_edges.append(e)
            filtered_genes = {}
            for gene, detail in drug_filtered_genes.items():
                if detail["is_seed"] or detail["has_edges_to"] or gene in all_edges:
                    filtered_genes[gene] = detail
            return {"drugs": resized_drugs, "genes": filtered_genes}

    if target == "drug-target":
        resized_genes = {}
        non_seed = []

        for gene, detail in genes.items():
            if detail["is_seed"]:
                resized_genes[gene] = detail
            else:
                non_seed.append(detail)

        if len(non_seed) <= result_size:
            for gene in non_seed:
                symbol = gene["symbol"]
                resized_genes[symbol] = gene

        if len(non_seed) > result_size:
            sorted_genes = sorted(non_seed, key=lambda item: item['score'])
            sorted_genes.reverse()
            for i in range(result_size):
                symbol = sorted_genes[i]["symbol"]
                resized_genes[symbol] = sorted_genes[i]
            for gene, detail in resized_genes.items():
                old_edges = detail["has_edges_to"]
                new_edges = []
                for edge in old_edges:
                    if edge in resized_genes.keys():
                        new_edges.append(edge)
                resized_genes[gene]["has_edges_to"] = new_edges

        return {"drugs": drugs, "genes": resized_genes}

    return {"drugs": drugs, "genes": genes}



"""
drugstone.scripts.wait_for_task_to_finish

This module implements the wait_for_task_to_finish function.

:copyright: 2022 Institute for Computational Systems Biology by Prof. Dr. Jan Baumbach
:author: Ugur Turhan
"""

import logging
import warnings
import time
from .request_task_info import request_task_info


def wait_for_task_to_finish(token: str, task_id: str = None) -> dict:
    """Waits for the task to finish and returns the task information."""

    info = request_task_info(token)
    while not info["done"] and not info["failed"]:
        logging.info(task_id + " progress is at: " + str(__get_progress(info)))
        time.sleep(2)
        info = request_task_info(token)
    if info["done"]:
        logging.info(task_id + " is done.")
    elif info["failed"]:
        warnings.warn(task_id + " has failed!")
    return info


def __get_progress(info: dict) -> str:
    f_progress = float(info["progress"]) * 100
    return str(f_progress) + "%"



"""
src.drugstone.scripts.fetch_edges

This module implements the fetch_edges function.

:copyright: 2022 Institute for Computational Systems Biology by Prof. Dr. Jan Baumbach
:author: Ugur Turhan
"""

import requests
from .constants.url import Url


def fetch_edges(internal_ids: list, ppi_dataset: str):

    n_ids = []

    for i in internal_ids:
        n_ids.append({"netex_id": i})

    data = {
        "nodes": n_ids,
        "dataset": ppi_dataset
    }

    return requests.post(
        Url.FETCH_EDGES,
        verify=False,
        json=data
    ).json()



"""
drugstone.task.tasks_result

This module implements the class :class:`TasksResult` for the drugstone API.

:copyright: 2022 Institute for Computational Systems Biology by Prof. Dr. Jan Baumbach
:author: Ugur Turhan
"""

import sys
import json
import logging
from typing import List
from .task import Task
from .scripts.create_path import create_path
from .scripts.make_upsetplot import make_upset_plot


class TasksResult:
    """Represents the results of a list of :class:`Task` objects.

    get_tasks_list() -> List[:class:`Task`]:
        Returns the list of tasks.

    to_dict() -> dict:
        Returns a dict with the results of the tasks.

    download_json(path: str, name: str) -> None:
        Downloads a json file with the results.

    create_upset_plot() -> None:
        Opens a new window with an upset plot of the results.
    """

    def __init__(self, tasks: List[Task] = []) -> None:
        self.__tasks = tasks

    def get_tasks_list(self) -> List[Task]:
        """Returns the list of tasks."""

        return self.__tasks

    def to_dict(self) -> dict:
        """Returns a dict with the results of the tasks."""

        d = {}
        for t in self.__tasks:
            d[t.get_parameters()["taskId"]] = {
                "info": t.get_info(),
                "parameters": t.get_parameters(),
                "results": t.get_result().to_dict()
            }
        return d

    def download_json(self, path: str = "", name: str = "result") -> None:
        """Downloads a json file with the results.

        :param str path: (optional) Path, where to download the file. Defaults to the current path.
        :param str name: (optional) Name for the file. Defaults to 'result'.
        """

        path = create_path(path, name, "json")
        with open(path, "x") as f:
            json.dump(self.to_dict(), f, indent=4)

    def create_upset_plot(self) -> None:
        """Opens a new window with an upset plot of the drug results.

        At least one of the tasks has to be a drug-search.
        This is only available with python 3.6!
        """

        has_drugs = False
        for task in self.get_tasks_list():
            if task.get_result().get_drugs():
                has_drugs = True

        if has_drugs:
            if sys.version_info[:2] == (3, 6):
                logging.info("IMPORTANT: The script pauses for the new window, for the UpSet plot! "
                             + "Close the UpSet plot window, for the script to continue or terminate! ")
                make_upset_plot(self.to_dict())
            else:
                logging.warn("create_upset_plot() is only compatible with Python 3.6!")
        else:
            logging.warn("Something went wrong! "
                         + "At least one task has to be a drug-search. "
                         + "No drugs were found.")



"""
drugstone.scripts.constants.url

This module implements the class Url.

:copyright: 2022 Institute for Computational Systems Biology by Prof. Dr. Jan Baumbach
:author: Ugur Turhan
"""


class Url:
    BASE = "https://api.drugst.one/"
    # BASE = "localhost:8001/"
    MAP_NODES = BASE + "map_nodes/"
    TASK = BASE + "task/"
    TASKS = BASE + "tasks/"
    TASK_RESULTS = BASE + "task_result/"
    FETCH_EDGES = BASE + "fetch_edges/"



"""
drugstone.scripts.constants.task_parameter

This module implements the class TaskParameter.

:copyright: 2022 Institute for Computational Systems Biology by Prof. Dr. Jan Baumbach
:author: Ugur Turhan
"""


class TaskParameter:

    # Available general parameters.
    IDENTIFIER = "identifier"
    ALGORITHM = "algorithm"
    ALGORITHMS = "algorithms"
    PPI = "ppi_dataset"
    PDI = "pdi_dataset"
    TARGET = "target"
    MAX_DEG = "max_deg"
    INCLUDE_INDIRECT_DRUGS = "include_indirect_drugs"
    HUB_PENALTY = "hub_penalty"
    RESULT_SIZE = "result_size"
    INCLUDE_NON_APPROVED_DRUGS = "include_non_approved_drugs"
    FILTER_PATHS = "filter_paths"

    # Available parameter for the trustrank algorithm.
    class Trustrank:
        DAMPING_FACTOR = "damping_factor"

    # Available parameter for the multi steiner algorithm.
    class MultiSteiner:
        NUM_TREES = "num_trees"
        TOLERANCE = "tolerance"
    
    # Available parameter for the keypathwayminer algorithm.
    class KeyPathwayMiner:
        K = "k"

    # Available options for the 'identifier' parameter.
    class IdentifierValues:
        IDENTIFIER_VALUES = ["symbol", "uniprot", "ensg"]
        SYMBOL = "symbol"
        UNIPROT = "uniprot"
        ENSG = "ensg"

    # Available options for the 'algorithm' parameter.
    class AlgorithmValues:
        DRUG_TARGET_SEARCH_VALUES = [
            "multisteiner", "keypathwayminer", "trustrank",
            "closeness", "degree", "betweenness"
        ]
        DRUG_SEARCH_VALUES = ["trustrank", "closeness", "degree", "proximity"]
        ALGORITHM_VALUES = DRUG_TARGET_SEARCH_VALUES + DRUG_SEARCH_VALUES
        MULTISTEINER = "multisteiner"
        KEYPATHWAYMINER = "keypathwayminer"
        TRUSTRANK = "trustrank"
        CLOSENESS = "closeness"
        DEGREE = "degree"
        PROXIMITY = "proximity"
        BETWEENNESS = "betweenness"
    
    # Available options for the 'ppi' parameter.
    class PpiValues:
        PPI_VALUES = ["STRING", "BioGRID", "APID", "NeDRex"]
        STRING = "STRING"
        BIOGRID = "BioGRID"
        APID = "APID"
    
    # Available options for the 'pdi' parameter.
    class PdiValues:
        PDI_VALUES = ["drugbank", "chembl", "dgidb", "NeDRex"]
        DRUGBANK = "drugbank"
        CHEMBL = "chembl"
        DGIDB = "dgidb"
    
    # Available options for the 'target' parameter.
    class TargetValues:
        TARGET_VALUES = ["drug", "drug-target"]
        DRUG = "drug"
        DRUG_TARGET = "drug-target"



"""
drugstone.new_tasks

This module implements the new_tasks function for the drugstone API.

:copyright: 2022 Institute for Computational Systems Biology by Prof. Dr. Jan Baumbach
:author: Ugur Turhan
"""

import warnings
from typing import Union, List
from .new_task import new_task
from .task.task import Task
from .task.tasks import Tasks
from .scripts.normalize_task_parameter import normalize_task_parameter


def new_tasks(
        seeds: list = list([]),
        parameters: Union[dict, List[dict]] = dict({}),
        static: bool = False,
        static_tasks: List[Task] = list([])) -> Tasks:
    """Starts multiple tasks.

    Starts multiple tasks, according to the user given seeds and parameters.
    Returns a :class:`Tasks` object, wrapping the individual the tasks.

    :param list seeds: (optional) List of seed nodes for the tasks.
    :param dict parameters: (optional) Dictionary of parameters for the tasks. Defaults to an empty dict {}.
    :param bool static: (optional)
    :param List[Task] static_tasks: (optional)
    :return: :class:`Tasks` object
    """

    # static list of tasks
    if static:
        return Tasks(static_tasks)

    # no seeds and no static
    if not static and not seeds:
        warnings.warn("Something went wrong! "
                      + "Maybe you forgot to pass seed genes, "
                      + "or forgot to set static to True.")
        return Tasks()

    # tasks with a list of parameter dictionaries
    if isinstance(parameters, list):
        algs = []
        for p in parameters:
            if "algorithm" in p:
                algs.append(p["algorithm"])
            elif "algorithms" in p:
                algs.append(p["algorithms"])
            else:
                algs.append(normalize_task_parameter({}, [])["algorithm"])
        has_duplicates = len(algs) != len(set(algs))
        tasks = []
        for p in parameters:
            new_p = {**p, "has_duplicate_algorithms": has_duplicates}
            t = new_task(seeds, new_p)
            tasks.append(t)
        return Tasks(tasks)

    # tasks with a list of algorithms
    if isinstance(parameters, dict):
        algorithms = None
        if "algorithm" in parameters:
            algorithms = parameters["algorithm"]
        elif "algorithms" in parameters:
            algorithms = parameters["algorithms"]
        if isinstance(algorithms, list):
            has_duplicates = len(algorithms) != len(set(algorithms))
            tasks = []
            for alg in algorithms:
                t_param = {
                    **parameters,
                    "algorithm": alg,
                    "has_duplicate_algorithms": has_duplicates
                }
                t_param.pop("algorithms", None)
                t = new_task(seeds, t_param)
                tasks.append(t)
            return Tasks(tasks)

    # Tasks with just a single task
    task = new_task(seeds, parameters)
    return Tasks([task])
